import 'package:organic_injector/src/errors/injector_error.dart';
import 'package:organic_injector/src/errors/injector_missing_mapping_error.dart';
import 'package:organic_injector/src/i_injector.dart';
import 'package:organic_injector/src/injectables/injectable.dart';
import 'package:organic_injector/src/mapping/injection_mapping.dart';
import 'package:organic_injector/src/providers/i_provider.dart';

class Injector implements IInjector {
  /*public fields*/
  static IInjector _instance;

  ///
  /// use this singleton if you want to use one unique injector across the app.
  /// It can help to have one class where you will map all necessary data.
  ///
  /// example :
  /// inside main() before you start the app call
  /// AppConfig();
  ///
  /// inside map your models or controllers
  ///
  /// AppConfig(){
  ///   Injector.instance.mapToValue(CounterNotifier,CounterNotifier(0);
  ///   Injector.instance.mapToValue(CounterModel,CounterModel());
  /// }
  ///
  static IInjector get instance {
    if (_instance == null) {
      _instance = new Injector();
    }
    return _instance;
  }

  Map<String, IProvider> providerMappings = new Map<String, IProvider>();
  Map<String, InjectionMapping> mappings = new Map<String, InjectionMapping>();

  Map<String, bool> _mappingsInProcess = new Map<String, bool>();
  List<dynamic> _managedObjects = new List<dynamic>();

  /// Parent injector used when you want to have different injectors with different scopes. Still not implemented fully
  IInjector parentInjector;

  InjectionMapping _map(Type type, [String name = '']) {
    final String mappingId = _getMappingId(type, name);

    if (mappings[mappingId] == null)
      return _createMapping(mappingId, type, name);
    else
      return mappings[mappingId];
  }

  InjectionMapping _createMapping(String mappingId, Type type, String name) {
    if (_mappingsInProcess[mappingId] != null) {
      throw new InjectorError(
          'Can\'t change a mapping from inside a listener to it\'s creation event');
    }

    _mappingsInProcess[mappingId] = true;

    final InjectionMapping mapping = new InjectionMapping(mappingId, this, type, name);

    mappings[mappingId] = mapping;

    final Symbol sealKey = mapping.seal();

    _mappingsInProcess.remove(mappingId);

    mapping.unseal(sealKey);

    return mapping;
  }

  static String _getMappingId(Type type, [String injectionName = '']) {
    return type.toString() + "|" + injectionName;
  }

  IProvider _getProvider(String mappingId) {
    Injector injector = this;
    IProvider provider = injector.providerMappings[mappingId];

    return provider;
  }

  /*public methods*/

  ///map value to Type. In case no name is provided, everytime you request [Type], you will get
  ///the same instance mapped to that [Type]. If you need new instance every time, consider creating [IProvider] and using [mapToProvider]
  ///
  /// by default, after mapping, object will be injected into with mapped dependecies
  void mapToValue(Type type, Object value,
      {String name: '', bool injectInto: true}) {
    _map(type, name).toValue(value);
    if (injectInto && value is IInjectable) {
      this.injectInto(value);
    }
  }

  ///map to a provider, check [IProvider]
  void mapToProvider(Type type, IProvider provider, {String name: ''}) {
    _map(type, name).toProvider(provider);
  }

  ///remove mapping from [Injector]
  void unmap(Type type, [String name = '']) {
    final String mappingId = _getMappingId(type, name);
    InjectionMapping mapping = mappings[mappingId];

    if ((mapping != null) && mapping.isSealed) {
      throw new InjectorError('can\'t unmap a sealed mapping');
    }
    if (mapping == null) {
      throw new InjectorError('Error while removing an injector mapping: ' +
          'No mapping defined for dependency ' +
          mappingId);
    }
    mapping.getProvider().destroy();
    mappings.remove(mappingId);
    providerMappings.remove(mappingId);
  }

  ///gets instance for the [type] and optional [name] based on previous mappings
  Object getInstance(Type type, {String name = ''}) {
    var object;

    final String mappingId = _getMappingId(type, name);
    IProvider provider = _getProvider(mappingId);
    if (provider == null)
      throw new InjectorMissingMappingError(
          'No mapping found for request ' + mappingId);
    else {
      object = provider.apply(type: type);
    }

    if (object is IInjectable) {
      injectInto(object);
    }

    return object;
  }

  ///populate dependencies of object that implements IInjectable
  void injectInto(IInjectable object) {
    object.inject(this);
  }

  ///Does Injector satisfies [type] and [name] dependecies
  bool satisfies(Type type, [String name = '']) {
    final String mappingId = _getMappingId(type, name);
    return _getProvider(mappingId) != null;
  }

  ///If you wish to manipulate mapping for any reason
  ///examples would be to seal the mapping, or change its provider.
  InjectionMapping getMapping(Type type, [String name = '']) {
    final String mappingId = _getMappingId(type, name);
    InjectionMapping mapping = mappings[mappingId];
    if (mapping == null) {
      throw new InjectorMissingMappingError(
          'Error while retrieving an injector mapping: ' +
              'No mapping defined for dependency ' +
              mappingId);
    }
    return mapping;
  }

  ///runs injector destroy process on the object if it includes InDestroy in its injectables
  void destroyInstance(IDestroy instance) {
    if (instance == null) return;

    instance.destroy();
  }

  /// Clears [Injector]
  void teardown() {
    mappings.forEach((String key, InjectionMapping mapping) {
      mapping.getProvider().destroy();
    });

    _managedObjects.forEach((dynamic instance) {
      destroyInstance(instance);
    });

    mappings = new Map<String, InjectionMapping>();
    _mappingsInProcess = new Map();
    _managedObjects = new List<dynamic>();
  }

  ///creates child injector. Still not implemented fully
  IInjector createChildInjector() {
    IInjector injector = new Injector();
    injector.parentInjector = this;
    return injector;
  }

  /// [bool] is there mappingId for [type] and [name]
  bool hasMapping(Type type, [String name = '']) {
    return _getProvider(_getMappingId(type, name)) != null;
  }
}
